#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>
#include <dirent.h>
#include <iconv.h>
#include <locale.h>
#include <langinfo.h>
#include <magic.h>
#include <libgen.h>
#include "mep.h"

int sock, error;

char *getid(char *host)
{
  char *id;

  id = (char *) malloc((20 + strlen(host)) * sizeof(char));
  sprintf(id, "%u.%u.%s", (int) getpid(), (int) time(NULL), host);

  return id;
}

char *getzone(long timezone, int isdst)
{
  char *zone = (char *) malloc(3 * sizeof(char));
  char *zones = "YXWVUTSRQPONZABCDEFGHIKLM??";
  int halfhours = 24 - timezone / 1800;

  zone[0] = zones[halfhours / 2 + (isdst > 0)];
  zone[1] = halfhours % 2 ? '+' : '\0';
  zone[2] = '\0';

  return zone;
}

char *gettime(void)
{
  char timestring[22], *zone;
  struct tm *local;
  time_t t;

  time(&t);
  local = localtime(&t);
  zone = getzone(timezone, local->tm_isdst);

  sprintf(timestring, "%04d-%02d-%02d %02d:%02d:%02d%s",
	  1900 + local->tm_year, 1 + local->tm_mon, local->tm_mday,
	  local->tm_hour, local->tm_min, local->tm_sec, zone);

  free(zone);

  return strdup(timestring);
}

void md5send(char *id, char *md5)
{
  char string[NI_MAXHOST + 53];
  int sock;
  struct sockaddr_un addr;

  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, UNIX_SOCKET);

  if ((sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "Unable to get a socket: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  if (connect(sock, (struct sockaddr *) &addr,
	      sizeof(struct sockaddr_un)) < 0) {
    fprintf(stderr, "Unable to connect: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  sprintf(string, "%s %s", id, md5);
  send(sock, string, strlen(string), 0);

  close(sock);
}

int parts(const struct dirent *list)
{
  char first;

  return (first = *(list->d_name)) > 0x30 && first < 0x40;
}

void freelist(struct dirent **namelist, int n)
{
  while (n--)
    free(namelist[n]);
  free(namelist);
}

void extract(int msg)
{
  char dir[PATH_MAX + 1], *temp;
  char from[PATH_MAX + 1], to[PATH_MAX + 1];
  struct dirent **namelist, **namelist2;
  int i, n, n2;

  if ((n = scandir(SPOOL_PATH, &namelist, 0, alphasort)) >= 0) {
    sprintf(dir, "%s/%s", SPOOL_PATH, namelist[msg + 1]->d_name);
    if ((n2 = scandir(dir, &namelist2, parts, alphasort)) >= 0) {
      for (i = 0; i < n2; i++) {
	printf("Extracting %s... ", namelist2[i]->d_name);

	if ((temp = strchr(namelist2[i]->d_name, '_')) != NULL) {
	  sprintf(from, "%s/%s", dir, namelist2[i]->d_name);

	  if (strlen(++temp) != 0)
	    sprintf(to, "./%s", temp);
	  else
	    sprintf(to, "./message%d", i);

	  if (rename(from, to) == 0)
	    printf("OK");
	}

	printf("\n");
      }

      sprintf(from, "%s/0_", dir);
      unlink(from);

      rmdir(dir);
    }

    freelist(namelist2, n2);
  }

  freelist(namelist, n);
}

void listspool(void)
{
  char dir[PATH_MAX + 1];
  struct dirent **namelist, **namelist2;
  int i, n, n2;

  if ((n = scandir(SPOOL_PATH, &namelist, 0, alphasort)) >= 0) {
    for (i = 0; i < n; i++) {
      if (strlen(namelist[i]->d_name) == 8) {
	printf("Message %d.\t%s", i - 1, namelist[i]->d_name);

	sprintf(dir, "%s/%s", SPOOL_PATH, namelist[i]->d_name);
	if ((n2 = scandir(dir, &namelist2, parts, alphasort)) >= 0)
	  printf("\t\t%d parts\n", n2);

	freelist(namelist2, n2);
      }
    }
  }

  freelist(namelist, n);
}

void check_reply(char *reply, char *message)
{
  char buffer[CHUNK_SIZE];

  buffer[recv(sock, buffer, CHUNK_SIZE, 0)] = '\0';

  if (strncmp(buffer, reply, strlen(reply)) != 0) {
    fprintf(stderr, message);
    close(sock);
    exit(EXIT_FAILURE);
  }
}

char *utf8_recode(char *string)
{
  char buffer[PATH_MAX + 1];
  char *outbuf = buffer;
  size_t insize, outsize = PATH_MAX + 1;
  iconv_t desc;

  insize = strlen(string);

  setlocale(LC_CTYPE, "");

  if ((desc = iconv_open("UTF-8", nl_langinfo(CODESET))) < 0) {
    fprintf(stderr, "Unable to get a conversion descriptor: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  iconv(desc, &string, &insize, &outbuf, &outsize);
  iconv_close(desc);

  *outbuf = '\0';

  return strdup(buffer);
}

char *mepify_mime(const char *string)
{
  char *new, *mime, *charset;

  new = strdup(string);
  mime = strsep(&new, "\t,;");

  if (new && (charset = strstr(new, "charset=")) != NULL) {
    if (strcmp(charset + 8, "us-ascii") == 0)
      strcpy(charset + 8, "utf-8");

    strcat(mime, ";");
    strcat(mime, charset);
  }

  if (strcmp(mime, "") == 0) {
    free(mime);
    mime = strdup("application/octet-stream");
  }

  return mime;
}

int main(int argc, char *argv[])
{
  unsigned char digest[16];
  char md5[32], buffer[CHUNK_SIZE];
  char host[NI_MAXHOST], *id;
  const char *magic;
  magic_t cookie;
  struct stat st;
  MD5_CTX context;
  int i, file, bytes;

  if (argc < 2) {
    listspool();
    exit(EXIT_SUCCESS);
  }

  if (argv[1][0] == '#') {
    extract(atoi(argv[1] + 1));
    exit(EXIT_SUCCESS);
  }

  if ((sock = getmxsocket(argv[1])) == 0) {
    fprintf(stderr, "Unable to connect to server\n");
    exit(EXIT_FAILURE);
  }

  check_reply("", "");

  if (gethostname(host, NI_MAXHOST) < 0)
    sprintf(host, "[unknown]");

  sprintf(buffer, "INIT MEP/0.9\r\n");
  send(sock, buffer, strlen(buffer), 0);

  check_reply("OK", "Incorrect reply from server\n");

  sprintf(buffer, "SEND HEAD %s\r\n", id = getid(host));
  send(sock, buffer, strlen(buffer), 0);

  check_reply("ACCE", "Message refused by server\n");

  sprintf(buffer, "Recipient: \r\nSender: \r\nSubject: \r\n");
  send(sock, buffer, strlen(buffer), 0);
  sprintf(buffer, "To: %s\r\nFrom: %s\r\n", argv[1], SENDER);
  send(sock, buffer, strlen(buffer), 0);
  sprintf(buffer, "Deliver: %s\r\nVerify: %s\r\n", argv[1], SENDER);
  send(sock, buffer, strlen(buffer), 0);
  sprintf(buffer, "Track: %s %s %s\r\n", gettime(), host, argv[1]);
  send(sock, buffer, strlen(buffer), 0);
  sprintf(buffer, "\r\n");
  send(sock, buffer, strlen(buffer), 0);

  check_reply("OK", "Message refused by server\n");

  MD5Init(&context);

  for (i = 2; i < argc; i++) {
    if (stat(argv[i], &st) < 0) {
      fprintf(stderr, "Unable to stat file %s: %s\n",
	      argv[i], strerror(errno));
      exit(EXIT_FAILURE);
    }

    if (!S_ISREG(st.st_mode)) {
      fprintf(stderr, "%s is not a regular file\n", argv[i]);
      exit(EXIT_FAILURE);
    }

    if ((cookie = magic_open(MAGIC_MIME)) == NULL ||
	magic_load(cookie, NULL) == -1 ||
	(magic = magic_file(cookie, argv[i])) == NULL) {
      fprintf(stderr, "%s\n", magic_error(cookie));
      exit(EXIT_FAILURE);
    }

    sprintf(buffer, "SEND PART %d %s %s\r\n", (int) st.st_size,
	    mepify_mime(magic), utf8_recode(basename(argv[i])));
    send(sock, buffer, strlen(buffer), 0);

    magic_close(cookie);

    check_reply("ACCE", "Message refused by server\n");

    fprintf(stderr, "Sending \"%s\"... ", argv[i]);

    if ((file = open(argv[i], O_RDONLY)) < 0) {
      fprintf(stderr, "Unable to open file %s: %s\n",
	      argv[i], strerror(errno));
      exit(EXIT_FAILURE);
    }

    while ((bytes = read(file, buffer, CHUNK_SIZE)) != 0) {
      send(sock, buffer, bytes, 0);
      MD5Update(&context, buffer, bytes);
    }

    close(file);

    check_reply("OK", "Message refused by server\n");

    fprintf(stderr, "OK\n");
  }

  MD5Final(digest, &context);
  for (i = 0; i < 16; i++)
    sprintf(md5 + 2 * i, "%02x", digest[i]);

  md5send(id, md5);
  free(id);

  sprintf(buffer, "QUIT\r\n");
  send(sock, buffer, strlen(buffer), 0);

  check_reply("OK", "Incorrect reply from server\n");

  return EXIT_SUCCESS;
}
